
%if 0

UNFILL - Free the UMBs allocated by my UMBFILL
Written by C. Masloch, 2008. Public Domain.

Assemble with NASM 2.03.01: NASM.EXE -fbin -oUNFILL.COM UNFILL.ASM
 (Some other NASM versions should also work)

UNFILL identifies UMBs filled by UMBFILL by first verifying if the MCB name is "UMBFILL",0.
If it's not, this is no block filled by UMBFILL. Then it's checked wether the MCB owner value
equals the actual block. If it does, then this a block filled by UMBFILL. If it doesn't, it's
last checked wether the MCB owner value equals 0008h or another block that points to itself
and also has the MCB name of "UMBFILL",0. If either condition matches, UMBFILL's PSP was
loaded here because UMBFILL was loaded with LH (and the UNFILL will free this block as well).

Report problems to <cm -AT- bttr-software.de> if you experience any and want them fixed by me.

%endif


		cpu 8086
		org 100h

start:
		push cs
		pop ds

		mov ax, 5802h			; get UMB link state
		int 21h
		jc error			; function not supported -->
		push ax				; save UMB link state

		mov ax, 5803h			; set UMB link state
		mov bx, 1			; enabled
		int 21h

		xor dx, dx			; dh = PSP flag. 0= scan for the UMBFILL PSP (special handling)
						; dl = End flag. 0= Current MCB is an 'M' MCB, 1= It's the 'Z' MCB
		xor cx, cx			; set UMB counter to 0

getfirstmcb:
		mov ah, 52h
		int 21h				; get DOS list of lists
		mov bx, [ es:bx - 2 ]		; get first MCB

		push cs
		pop es

nextmcb:		; scan in next MCB
			; bx = Next MCB, dx = search flags, cx = number of already freed UMBs
		push ds
		mov ds, bx

		cmp byte [ 0000h ], 'M'		; not last MCB ?
		je .valid			; yes -->

		inc dl				; indicate last MCB
		cmp byte [ 0000h ], 'Z'		; valid MCB ?
		jne error			; no -->

.valid:
		cmp bx, 9FFFh			; is it an UMB at all ?
		jb .next			; no, just walk to next -->

		call checkmcb
		jc .next			; doesn't match -->

		inc cx				; found matching, increase counter

		push es
		mov ax, ds
		inc ax
		mov es, ax			; must pass block segment to DOS
		mov ah, 49h			; free memory
		int 21h
		pop es

.next:
		add bx, word [ 0003h ]		; add size of this block
		inc bx				; add one to get next PSP
		pop ds
		cmp dl, 0			; last MCB was 'Z' ?
		je nextmcb			; no, continue until 'Z' MCB hit
		xor dl, dl			; reset to 0
		cmp dh, 1			; Only PSP search completed ? (if below, it's zero)
		inc dh				; indicate to search normal UMBFILL blocks now
						; (inc doesn't affect CF)
		jb getfirstmcb			; yes, start normal search from first MCB -->

checkifany:
		jcxz error			; no matching UMBs found at all ? -->

		mov dx, msg.done
		mov ah, 09h
		int 21h
		
		int 20h				; terminate


error:
		mov dx, msg.noumbs
		mov ah, 09h
		int 21h

		int 20h


checkmcb:		; check if MCB is an UMBFILL MCB
			; es = cs, ds = checked MCB
			; dh = 0 if UMBFILL PSP special check, = 1 if normal check
			;
			; Returns CY if no UMBFILL MCB, NC else
			;
			; Uses di, si and ax
		cmp word [ 0001h ], byte 0000h	; already free ?
		je .no				; yes, then no match -->

		mov di, msg.umbfill
		mov si, 0008h

		push cx
		mov cx, si			; Length is same as starting address
		repe cmpsb			; does the MCB name match ?
		pop cx
		jne .no				; no -->

		cmp dh, 0
		je .psp
.normal:
		mov ax, ds
		inc ax
		cmp ax, word [ 0001h ]		; owner references block itself ?
		jne .no				; no -->
						; else: match. flags: ZR, NC
		retn

.psp:
		call .normal
		jnc .no				; mustn't point to itself -->
		mov ax, word [ 0001h ]
		cmp ax, byte 0008h		; marked as owned by DOS ?
		je .yes				; yes, matches -->

		push ds
		push dx
		dec ax
		mov ds, ax			; get the MCB of the owner block
		inc dh				; indicate to check normal now
		call checkmcb			; check if it's ok
		pop dx
		pop ds
						; leave CF
.yes:
		retn

.no:
		stc
		retn


msg:
.umbfill:	db "UMBFILL",0			; searched MCB name
.noumbs:	db "No UMBFILL UMBs found",13,10,36
.done:		db "All UMBFILL UMBs deallocated",13,10,36
